//
//  main.c
//
//  VA_Pmeter 0.5 / Digital Volt+Ampare panel meter 
//  Copyright (C) 2010 Toyohiko Togashi tog001@nifty.com
//
//  This program is free software; you can redistribute it and/or modify it under the terms of the
//  GNU General Public License as published by the Free Software Foundation; either version 3
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along with this program.
//  If not, see <http://www.gnu.org/licenses/>
//
//
//	Description
//		Digital panel meter.
//		Voltage and Current or Power with warning sound.
//		Using 8x2 character LCD module "SC0802".
//
//	Setting
//		PIC16
//			Device	: PIC16F88
//			IDE		: MPLAB IDE v8.36,
//			          HI-TECH C PRO for the PIC10/12/16 MCU Family(Lite) V9.65PL1
//
//	Device configuration
//		Clock:	4MHz/Internal
//		Internal peripheral module:
//			WDT:	-
//			BOR:	-
//			EEPROM:	Use
//			Timer0:	-
//			Timer1:	-
//			Timer2:	-
//			CCP:	-
//			SSP:	-
//			AUSART:	-
//			ADC:	Voltage & current measure
//			Comp:	-
//			Vref:	-
//		Pin assign:
//			1:	RA2/AN2		VIx1	
//			2:	RA3/AN3		VIx1001
//			3:	RA4/AN4		VIx101
//			4:	RA5/MCLR	SW-A
//			5:	Vss			GND
//			6:	RB0/CCP1	LCD-SC0802-DB4
//			7:	RB1/SDI/SDA	LCD-SC0802-DB5
//			8:	RB2/SDO/RX	LCD-SC0802-DB6
//			9:	RB3/PGM		LCD-SC0802-DB7
//			10:	RB4/SCK/SCL	LCD-SC0802-RS
//			11:	RB5/SS/TX	LCD-SC0802-R/W
//			12:	RB6/AN5/PGC	LCD-SC0802-E
//			13:	RB7/AN6/PGD	Buzzer
//			14:	Vdd			+5V
//			15:	RA6/OSC2	SW-B
//			16:	RA7/OSC1	SW-C
//			17:	RA0/AN0		VV/1
//			18:	RA1/AN1		VV/11
//
//	Usage
//		Push button A: Reset
//		Push button B: Enter setup mode  / setup->next values
//		Push button C: Watt display mode / setup->change value
//
//
//	Update history
//	---------- ----- -------------------------------------------
//	2010.12.08 v0.5  1st release
//
//

#include <stdio.h>
#include "htc.h"						// HI-TECH C
#include "eeprom_routines.h"			// HI-TECH C
#include "lcd.h"						// Character LCD support

#define SW_A		(!RA5)
#define SW_B		(!RA6)
#define SW_C		(!RA7)

#define AD_VREF			5000			// mV
#define AD_FULLSCALE	1024			// Count
#define V_RANGE			2				// Voltage channel #
#define I_RANGE			3				// Current channel #

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define ABS(a)		(((a) >  0 ) ? (a) : -(a))

//---------------------------------------------------------------------
//
//	Config
//
//---------------------------------------------------------------------
//	CPU Configuration Fuses bits
__CONFIG(0x3fff

// Protection of program code 
//& PROTECT
& UNPROTECT

// CCP1 Pin selection 
& CCPRB0
//& CCPRB3

// In-Circuit Debugger Mode 
//& DEBUGEN
& DEBUGDIS

// Flash Program Memory Write Enable 
& UNPROTECT
//& WP0
//& WP1
//& WPA

// Data EE Memory Code Protection 
& UNPROTECT
//& CPD

// Low Voltage Programming Enable 
//& LVPEN
& LVPDIS

// Brown out detection enable 
//& BOREN
& BORDIS

// Master clear reset pin function
& MCLREN
//& MCLRDIS

// Power up timer enable 
& PWRTDIS
//& PWRTEN

// Watchdog timer enable 
//& WDTEN
& WDTDIS

// Oscillator configurations 
//& RCCLK
//& RCIO
//& INTCLK
& INTIO
//& EC
//& HS
//& XT
//& LP
);

//	EEPROM initial data
__EEPROM_DATA(0, 0, 0, 0, 0, 0, 0, 0);
__EEPROM_DATA(0, 0, 0, 0, 0, 0, 0, 0); 
__EEPROM_DATA(0, 0, 0, 0, 0, 0, 0, 0);
__EEPROM_DATA(0, 0, 0, 0, 0, 0, 0, 0);
__EEPROM_DATA(0, 0, 0, 0, 0, 0, 0, 0);
__EEPROM_DATA(0, 0, 0, 0, 0, 0, 0, 0);

//
// Connect stdio -> LCD
//
void putch(char c) {
	LCD_WriteData(c);
}

//---------------------------------------------------------------------
//
// Read A/D channel
//
//---------------------------------------------------------------------
unsigned short getad(char ch) {
	ADCON0 = (ADCON0 & 0b11000111) | (ch << 3);
	ADIF = 0;
	ADGO = 1;
	while(!ADIF);
	return((ADRESH << 8) | ADRESL);
}
unsigned short getads(char ch, char samp) {
	long s;
	signed char i;

	s = 0;
	for(i = samp; i > 0; --i) {
		s += getad(ch);
		__delay_ms(1);
	}
	return(s / samp);
}

//---------------------------------------------------------------------
//
// Sound
//
//---------------------------------------------------------------------
void beep(void){
	signed char	i;

	for(i = 100; i > 0; --i) {
		PORTB ^= 0b10000000;
		__delay_ms(1);
	}
}
void click(void){
	signed char	i;

	for(i = 5; i > 0; --i) {
		PORTB ^= 0b10000000;
		__delay_ms(1);
	}
}

//---------------------------------------------------------------------
//
//	EEPROM I/O
//
//---------------------------------------------------------------------
void eeprom_writes(unsigned char addr, unsigned char *buf, unsigned char len){
	for(; len > 0; --len) {
		EEPROM_WRITE(addr++, *buf++);
	}
	return;
}
void eeprom_reads(unsigned char addr, unsigned char *buf, unsigned char len){
	for(; len > 0; --len) {
		*buf++ = EEPROM_READ(addr++);
	}
	return;
}

//---------------------------------------------------------------------
//
//  M a i n
//
//---------------------------------------------------------------------
main(void){
	static struct {
		char	v_ch[V_RANGE];			// A/D ANx
		short	v_gain_a[V_RANGE];		// Terminal->Pin a/b
		short	v_gain_b[V_RANGE];		// Terminal->Pin a/b
		char	i_ch[I_RANGE];			// A/D ANx
		short	i_gain_a[I_RANGE];		// Terminal->Pin a/b
		short	i_gain_b[I_RANGE];		// Terminal->Pin a/b
		short	i_offset[I_RANGE];		// Zero adjust A/D count
		short	v_max;					// 1=0.001[V]
		short	v_delta;				// 1=0.001[V]
		short	i_max;					// 1=0.1[mA]
		short	i_delta;				// 1=0.1[mA]
		short	ad_autorange_v;			// A/D count
		short	ad_autorange_i;			// A/D count
		char	ad_sample_times;		//
	} ee;
	char	disp_mode = 0;				// 0:VA 1:VW


//
//		*** System clock ***
//
//				  OSCCON - Oscillator control register
//				  -0000000  Initial
//				  .XXX....  IRCF: Internal oscillator
//				  .000....     31.25KHz
//				  .001....     125KHz
//				  .010....     250KHz
//				  .011....     500KHz
//				  .100....     1MHz
//				  .101....     2MHz
//				  .110....     4MHz
//				  .111....     8MHz
//				  ....X...  OSTS: Oscillator Start-up Time-out Status bit
//				  ....0...     Device is running from T1OSC or INTRC as secondary system clock
//				  ....1...     Device is running from the primery system clock
//				  .....X..  IOFS:INTOSC Frequency stable bit
//				  .....1..     stable
//				  .....0..     not stable
//				  ......XX  SCS: Oscillator Mode Select bits
//				  ......00    Oscillator mode defined by FOSC<2:0>
//				  ......01    T1OSC is used by system clock
//				  ......10    Internal RC is used for system clock
//				  ......11    Reserved
#if _XTAL_FREQ == 31250					//
	OSCCON =    0b00000000;				// Internal 31.25KHz
#elif _XTAL_FREQ == 125000				//
	OSCCON =    0b00010000;				// Internal 125KHz
#elif _XTAL_FREQ == 250000				//
	OSCCON =    0b00100000;				// Internal 250KHz
#elif _XTAL_FREQ == 500000				//
	OSCCON =    0b00110000;				// Internal 500KHz
#elif _XTAL_FREQ == 1000000				//
	OSCCON =    0b01000000;				// Internal 1MHz
#elif _XTAL_FREQ == 2000000				//
	OSCCON =    0b01010000;				// Internal 2MHz
#elif _XTAL_FREQ == 4000000				//
	OSCCON =    0b01100000;				// Internal 4MHz
#elif _XTAL_FREQ == 8000000				//
	OSCCON =    0b01110000;				// Internal 8MHz
#else									//
#error "_XTAL_FREQ value"				//
#endif									//
//				  OSCTUNE - Oscillator tuning register
//				  00000000  Initial
//				  ..XXXXXX  TUN: Frequency tuning bits
//				  ..011111     +31	+12.5%
//				  ..000001     +1	+0.4%
//				  ..000000     0
//				  ..111111     -1	-0.4%
//				  ..100000     -32	-12.5%
	OSCTUNE = 0;
//
//		*** System option ***
//
//				  OPTION_REG - Option register
//				  11111111  Initial
//				  X.......  RBPU: PORTB Pull-up enable
//				  1.......     disabled
//				  0.......     enable
//				  .X......  INTEDG: Interrupt edge select
//				  .1......     riging edge
//				  .0......     failling edge
//				  ..X.....  T0CS: TMR0 clock source select
//				  ..1.....     Transition on GP2/T0CKI pin
//				  ..0.....     Internal instruction cycle
//				  ...X....  T0SE: TMR0 source edge select
//				  ...1....     high-to-low edge
//				  ...0....     low-to-high edge
//				  ....X...  PSA: Prescaler assignment
//				  ....1...     WDT
//				  ....0...     Timer0
//				  .....XXX  PS:  Prescaler assignment
//				  .....000     TMR0 rate 1:2 / WDT rate 1:1
//				  .....001               1:4 /          1:2
//				  .....010               1:8 /          1:4
//				  .....011              1:16 /          1:8
//				  .....100              1:32 /         1:16
//				  .....101              1:64 /         1:32
//				  .....110             1:128 /         1:64
//				  .....111             1:256 /        1:128
	OPTION =    0b11010100;		// Timer0=CLKO(4MHz/4)/32 -> 31.25KHz
//
//		*** Watchdock timer ***
//
//				  WDTCON - Watchdog control register
//				  ---0100?  Initial
//				  ...XXXX.  WDTPS: Watchdog timer period select bits
//				  ...0000.     1:32
//				  ...0001.     1:64
//				  ...0010.     1:128
//				  ...0011.     1:256
//				  ...0100.     1:512
//				  ...0101.     1:1024
//				  ...0110.     1:2048
//				  ...0111.     1:4096
//				  ...1000.     1:8192
//				  ...1001.     1:16384
//				  ...1010.     1:32768
//				  ...1011.     1:65536
//				  .......X  WDTPS: Watchdog timer period select bits
//				  .......1     WDT is turned on
//				  .......0     WDT is turned off
	WDTCON =    0b00010100;		// <WDT disable>
//	WDTCON =    0b00010101;		// 32.25KHz / 32768 ~= 1sec  <enable>
//	WDTCON =    0b00010111;		// 32.25KHz / 65536 ~= 2sec  <enable>
//
//		*** A/D Converter ***
//
//				  ADCON0 - A/D control register 0
//				  00000000  Initial
//				  XX......  ADCS: A/D converter clock select
//				  00......    ADCS2=0 Fosc/2   ADCS2=1 Fosc/4
//				  01......            Fosc/8           Fosc/16
//				  10......            Fosc/32          Fosc/64
//				  11......            FRC              FRC
//				  ..XXX...  CHS: Analog channel select bits
//				  ..000...    Channel 0 RA0/AN0
//				  ..001...    Channel 1 RA1/AN1
//				  ..010...    Channel 2 RA2/AN2
//				  ..011...    Channel 3 RA3/AN3
//				  ..100...    Channel 4 RA4/AN4
//				  ..101...    Channel 5 RB6/AN5
//				  ..110...    Channel 6 RB7/AN6
//				  .....X..  GO/DONE: A/D conversion status bit
//				  .....1..    ADON=1 in progress
//				  .....0..           not in progress
//				  .......X  ADON: A/D on bit
//				  .......1    module is operating
//				  .......0    module is shutoff
//	ADCON0 =    0b10000001;		// Fosc/2 <Channel dynamic change>
//	ADCON0 =    0b01000001;		// Fosc/8 <Channel dynamic change>
	ADCON0 =    0b11000001;		// FRC
//
//				  ADCON1 - A/D control register 1
//				  00000000  Initial
//				  X.......  ADFM: A/D result format select bit
//				  1.......    Right justified
//				  0.......    Left justified
//				  .X......  ADCS2: A/D clock divied by 2 select bit
//				  .1......    Enabled
//				  .0......    Disabled
//				  ..XX....  VCFG: A/D port configuration control bits
//				  ..00....    VREF+=AVdd  VREF-=AVss
//				  ..01....    VREF+=AVdd  VREF-=VREF-
//				  ..10....    VREF+=VREF+ VREF-=AVss
//				  ..11....    VREF+=VREF+ VREF-=VREF-
	ADCON1 =    0b10000000;		// Right adj, clock/1, Vref=Vdd:Vss
//	ADCON1 =    0b11000000;		// Right adj, clock/2, Vref=Vdd:Vss
//
//				  ANSEL - Analog select register
//				  01111111  Initial
//				  .X......  ANS6 1=Analog 0=Digital
//				  ..X.....  ANS5 1=Analog 0=Digital
//				  ...X....  ANS4 1=Analog 0=Digital
//				  ....X...  ANS3 1=Analog 0=Digital
//				  .....X..  ANS2 1=Analog 0=Digital
//				  ......X.  ANS1 1=Analog 0=Digital
//				  .......X  ANS0 1=Analog 0=Digital
	ANSEL =     0b00011111;		// 
//
//		*** GPIO port ***
//
//				  PORTA - General purpose I/O register A
//				  xxx00000  Initial
//				  X.......  RA7: OSC1/CLKI
//				  .X......  RA6: OSC2/CLKO
//				  ..X.....  RA5: MCLR/Vpp
//				  ...X....  RA4: AN4/T0CKI
//				  ....X...  RA3: AN3/Vref+
//				  .....X..  RA2: AN2/Vref-
//				  ......X.  RA1: AN1
//				  .......X  RA0: AN0
	PORTA =		0b11111111;		//
//
//				  PORTB - General purpose I/O register B
//				  xxxxxxxx  Initial
//				  X.......  RB7: AN6/T1OSI/PGD
//				  .X......  RB6: AN5/T1OSO/T1CKI/PGC
//				  ..X.....  RB5: SS/TX/CK
//				  ...X....  RB4: SCK/SCL
//				  ....X...  RB3: CCP1/PGM
//				  .....X..  RB2: SDO/RX/DT
//				  ......X.  RB1: SDI/SDA
//				  .......X  RB0: INT/CCP
	PORTB =		0b11111111;		//
//
//				  TRISA - GPIO tri-state register A
//				  11111111  Initial
//				  1.......  TRISA7: RA7 1=input 0=output
//				  .0......  TRISA6: RA6 1=input 0=output
//				  ..1.....  TRISA5: RA5 1=input only
//				  ...X....  TRISA4: RA4 1=input 0=output
//				  ....X...  TRISA3: RA3 1=input 0=output
//				  .....X..  TRISA2: RA2 1=input 0=output
//				  ......X.  TRISA1: RA1 1=input 0=output
//				  .......X  TRISA0: RA0 1=input 0=output
	TRISA =		0b10111111;		//
//
//				  TRISB - GPIO tri-state register B
//                 11111111  Initial
//				  X.......  TRISB7: RB7 1=input 0=output
//				  .X......  TRISB6: RB6 1=input 0=output
//				  ..X.....  TRISB5: RB5 1=input 0=output
//				  ...X....  TRISB4: RB4 1=input 0=output
//				  ....X...  TRISB3: RB3 1=input 0=output
//				  .....X..  TRISB2: RB2 1=input 0=output
//				  ......X.  TRISB1: RB1 1=input 0=output
//				  .......X  TRISB0: RB0 1=input 0=output
	TRISB =		0b00000000;		//

//		*** LCD module ***
	LCD_Init();
	LCD_Start();

	//
	// Recall parameter
	//

	// Get data
	eeprom_reads(0, (char *)&ee, sizeof(ee));
	if (ee.ad_sample_times == 0) {		// No data, init values
		ee.v_ch[0]     = 0;				// AN0
		ee.v_gain_a[0] = 1;				// Direct in
		ee.v_gain_b[0] = 1;				//
		ee.v_ch[1]     = 1;				// AN1
		ee.v_gain_a[1] = 1;				// 
		ee.v_gain_b[1] = 11;			// register=10:1
		ee.i_ch[0]     = 3;				// AN3
		ee.i_gain_a[0] = 1001;			// OP-AMP gain
		ee.i_gain_b[0] = 20;			// Rsense=0.1ohm: x 10  AntiMinusOffsetMixer: x 2
		ee.i_offset[0] = 0;				//
		ee.i_ch[1]     = 4;				// AN4
		ee.i_gain_a[1] = 101;			// OP-AMP gain
		ee.i_gain_b[1] = 20;			// Rsense=0.1ohm: x 10  AntiMinusOffsetMixer: x 2
		ee.i_offset[1] = 0;				//
		ee.i_ch[2]     = 2;				// AN2
		ee.i_gain_a[2] = 1;				// No amp
		ee.i_gain_b[2] = 20;			// Rsense=0.1ohm: x 10  AntiMinusOffsetMixer: x 2
		ee.i_offset[2] = 0;				//
		ee.v_max       = 20000;			// 20V
		ee.v_delta     = 100;			// 0.1V
		ee.i_max       = 7000;			// 700mA
		ee.i_delta     = 10;			// 1mA
		ee.ad_sample_times = 20;		//
		ee.ad_autorange_v  = 1000;		// 1000/1024
		ee.ad_autorange_i  = 700;		// 700/1024  OP-AMP output width=3.5V/5V
	}

	//
	// Main loop
	//
	for(;;) {
		char	v_range;				// Measure V range#
		char	i_range;				// Measure A range#
		short	v_mv;					// 1=0.001[V]
		short	v_mv_b;					// 1=0.001[V]
		short	i_100ua;				// 1=0.1[mA]
		short	i_100ua_b;				// 1=0.1[mA]
		short	a;						// Temporary work area

		//
		// Measure
		//

		// Voltage
		for(v_range = 0; v_range < V_RANGE; v_range++) {
			a = getads(ee.v_ch[v_range], ee.ad_sample_times);
			if (a < ee.ad_autorange_v) break;
		}
		v_mv = (long)a * AD_VREF * ee.v_gain_b[v_range] / AD_FULLSCALE / ee.v_gain_a[v_range];
		v_mv = v_mv - (i_100ua / 100);	// Rsense=0.1ohm

		// Current
		for(i_range = 0; i_range < I_RANGE; i_range++) {
			a = getads(ee.i_ch[i_range], ee.ad_sample_times);
			if (a < ee.ad_autorange_i) break;
		}
		i_100ua = (long)(a - ee.i_offset[i_range]) * AD_VREF * ee.i_gain_b[i_range] * 10 / AD_FULLSCALE / ee.i_gain_a[i_range];

		//
		// Display
		//
		LCD_Position(0,0);
		switch(v_range) {
		case 0:
			a = v_mv + 5;
			printf("%4d.%02dV", a/1000, (a%1000)/10);
			break;
		case 1:
			a = v_mv + 50;
			printf("%5d.%dV", a/1000, (a%1000)/100);
			break;
		default:
			break;
		}
		LCD_Position(1,0);
		if (disp_mode == 0) {
			switch(i_range) {
			case 0:
//				a = MAX(i_100ua, 0);
				a = i_100ua;
				printf("%4d.%dmA", a/10, a%10);
				break;
			case 1:
				a = i_100ua + 5;
				printf("%6dmA", a/10);
				break;
			case 2:
				a = i_100ua + 500;
				printf("%5d.%dA", a/10000, (a%10000)/1000);
				break;
			default:
				break;
			}
		} else {
			a = (long)v_mv * (long)i_100ua / 10000;
			printf("%6dmW", a);
		}

		//
		// Check values
		//

		// Limit
		if ((v_mv > ee.v_max) || (i_100ua > ee.i_max))  {
			beep();
		}

		// Difference
		if ((ee.v_delta > 0) &&	((ABS(v_mv - v_mv_b) > ee.v_delta))) {
			click();
		}
		if ((ee.i_delta > 0) && ((ABS(i_100ua - i_100ua_b) > ee.i_delta))) {
			click();
		}
		v_mv_b    = v_mv;
		i_100ua_b = i_100ua;

		//
		// Button
		//
		if (SW_B) {						// Setup switch
			char	m;

			for(m = 0; ; ) {

				// Display menu & values
				LCD_Position(0, 0);
				switch(m) {
				case 0:
					printf("ADJ-ZERO");
					break;
				case 1:
					printf("ADJ-V   ");
					break;
				case 2:
					printf("ADJ-A   ");
					break;
				case 3:
					printf("MAX-V   ");
					break;
				case 4:
					printf("MAX-A   ");
					break;
				case 5:
					printf("DELTA-V ");
					break;
				case 6:
					printf("DELTA-A ");
					break;
				case 7:
					printf("INIT    ");
					break;
				default:
					break;
				}
				LCD_Position(1, 0);
				switch(m) {
				case 0:
					printf("%8d", ee.i_offset[0]);
					break;
				case 1:
					printf("N/A     ");
					break;
				case 2:
					printf("N/A     ");
					break;
				case 3:
					printf("%6dmV", ee.v_max);
					break;
				case 4:
					printf("%6dmA", ee.i_max/10);
					break;
				case 5:
					printf("%6dmV", ee.v_delta);
					break;
				case 6:
					printf("%6dmA", ee.i_delta/10);
					break;
				case 7:
					printf("OK?     ");
					break;
				default:
					break;
				}

				// Wait for push
				while((SW_B) || (SW_C));
				while((!SW_B) && (!SW_C));
				click();

				// Action
				if (SW_B) {
					m++;
					if (m > 7) {
						// Save EEPROM
						eeprom_writes(0, (char *)&ee, sizeof(ee));
						LCD_Position(0, 0);
						printf("        ");
						LCD_Position(1, 0);
						printf("SAVED   ");
						beep();
						__delay_ms(100);
						break;
					}
				}
				if (SW_C) {
					unsigned char	i;

					switch(m) {
					case 0:
						for(i = 0; i < I_RANGE; i++) {
							ee.i_offset[i] = getads(ee.i_ch[i], ee.ad_sample_times);
						}
						break;
					case 3:
						ee.v_max += 1000;
						if (ee.v_max > 20000) {
							ee.v_max = 1000;
						}
						break;
					case 4:
						ee.i_max += 1000;
						if (ee.i_max > 20000) {
							ee.i_max = 1000;
						}
						break;
					case 5:
						ee.v_delta += 100;
						if (ee.v_delta > 1000) {
							ee.v_delta = 0;
						}
						break;
					case 6:
						ee.i_delta += 100;
						if (ee.i_delta > 1000) {
							ee.i_delta = 0;
						}
						break;
					case 7:
						ee.ad_sample_times = 0;
						break;
					default:
						break;
					}
				}
			}
		}
		if (SW_C) {						// Display mode switch
			disp_mode = 1;
		}

		//
		// Wait
		//
		__delay_ms(100);

	}
}
